"""Module for sampling from a distribution of regression environments."""

from collections import OrderedDict
from functools import partial

import sklearn.datasets

from .data_environment import DataEnvironment
from ..data_types import FeatureType, TargetType, DataSourceType


def envs(n=5, test_size=None, random_state=None, verbose=False):
    # TODO: limit number of envs by n
    return OrderedDict([
        ("sklearn.boston", DataEnvironment(
            name="sklearn.boston",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.REGRESSION,
            feature_types=[
                FeatureType.CONTINUOUS,  # CRIM
                FeatureType.CONTINUOUS,  # ZN
                FeatureType.CONTINUOUS,  # INDUS
                FeatureType.CATEGORICAL,  # CHAS
                FeatureType.CONTINUOUS,  # NOX
                FeatureType.CONTINUOUS,  # RM
                FeatureType.CONTINUOUS,  # AGE
                FeatureType.CONTINUOUS,  # DIS
                FeatureType.CONTINUOUS,  # RAD
                FeatureType.CONTINUOUS,  # TAX
                FeatureType.CONTINUOUS,  # PTRATIO
                FeatureType.CONTINUOUS,  # B
                FeatureType.CONTINUOUS,  # LSTAT
            ],
            feature_indices=range(13),
            fetch_training_data=partial(
                sklearn.datasets.load_boston, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
        ("sklearn.diabetes", DataEnvironment(
            name="sklearn.diabetes",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.REGRESSION,
            feature_types=[FeatureType.CONTINUOUS for _ in range(10)],
            feature_indices=range(10),
            fetch_training_data=partial(
                sklearn.datasets.load_diabetes, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
        ("sklearn.linnerud", DataEnvironment(
            name="sklearn.linnerud",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.MULTIREGRESSION,
            feature_types=[FeatureType.CONTINUOUS for _ in range(3)],
            feature_indices=range(3),
            fetch_training_data=partial(
                sklearn.datasets.load_linnerud, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
    ])



"""Data Preprocessor components."""

from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import (
    QuantileTransformer, OneHotEncoder, MinMaxScaler,
    StandardScaler, RobustScaler, Normalizer)
from sklearn.impute import SimpleImputer

from .algorithm_component import AlgorithmComponent
from .hyperparameter import (
    CategoricalHyperparameter, UniformIntHyperparameter,
    UniformFloatHyperparameter, TuplePairHyperparameter,
    BaseEstimatorHyperparameter, EmbeddedEstimatorHyperparameter)
from ..data_types import AlgorithmType


def simple_imputer():
    """Create a categorical and numeric imputer."""
    # this is a placeholder index for which columns should be transformed is
    # task environment specific.

    def init_simple_imputer(
            component_class, categorical_features, numeric_features):

        return component_class(transformers=[
            ("continuous_imputer", SimpleImputer(), numeric_features),
            ("categorical_imputer", SimpleImputer(), categorical_features),
        ])

    return AlgorithmComponent(
        name="SimpleImputer",
        component_class=ColumnTransformer,
        component_type=AlgorithmType.IMPUTER,
        initialize_component=init_simple_imputer,
        hyperparameters=[
            EmbeddedEstimatorHyperparameter(
                "continuous_imputer",
                CategoricalHyperparameter(
                    "strategy",
                    ["mean", "median", "most_frequent"],
                    default="mean")),
            EmbeddedEstimatorHyperparameter(
                "continuous_imputer",
                CategoricalHyperparameter(
                    "add_indicator", [True, False], default=True)),
            EmbeddedEstimatorHyperparameter(
                "categorical_imputer",
                CategoricalHyperparameter(
                    "strategy", ["most_frequent"],
                    default="most_frequent")),
            EmbeddedEstimatorHyperparameter(
                "categorical_imputer",
                CategoricalHyperparameter(
                    "add_indicator", [True, False], default=True)),
        ],
        constant_hyperparameters={"remainder": "passthrough"}
        )


def one_hot_encoder():
    """Create a one hot encoder component.

    DeepCASH assumes explicit handling of which features are categorical and
    which are continuous, so this component would only be applied to the
    categorical features of a particular dataset.

    The categorical_features hyperparameter is set as
    data-environment-dependent because the controller isn't designed to
    propose the feature columns in which to perform feature transformations.
    This may change in the future. By definition, data-environment-dependent
    hyperparameters should be supplied as an element in the `hyperparameters`
    AlgorithmComponent.__init__ argument.

    For more details, see:
    http://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html
    """

    def init_one_hot_encoder(
            component_class, categorical_features, numeric_features):
        return component_class(transformers=[
            ("numeric_features", "passthrough", numeric_features),
            ("one_hot_encoder", OneHotEncoder(), categorical_features),
        ])

    return AlgorithmComponent(
        name="OneHotEncoder",
        component_class=ColumnTransformer,
        component_type=AlgorithmType.ONE_HOT_ENCODER,
        initialize_component=init_one_hot_encoder,
        constant_hyperparameters={
            "remainder": "passthrough",
            "one_hot_encoder__drop": None,
            "one_hot_encoder__handle_unknown": "ignore",
            "one_hot_encoder__sparse": False,
            "one_hot_encoder__categories": "auto",
        })


def minmax_scaler():
    """Create a minmax scaler component."""
    return AlgorithmComponent(
        name="MinMaxScaler",
        component_class=MinMaxScaler,
        component_type=AlgorithmType.RESCALER)


def standard_scaler():
    """Create a standard scaler component."""
    return AlgorithmComponent(
        name="StandardScaler",
        component_class=StandardScaler,
        component_type=AlgorithmType.RESCALER,
        hyperparameters=[
            CategoricalHyperparameter(
                "with_mean", [True, False], default=True),
            CategoricalHyperparameter(
                "with_std", [True, False], default=True),
        ])


def robust_scaler():
    """Create a robust scaler component."""
    return AlgorithmComponent(
        name="RobustScaler",
        component_class=RobustScaler,
        component_type=AlgorithmType.RESCALER,
        hyperparameters=[
            CategoricalHyperparameter(
                "with_centering", [True, False], default=True),
            CategoricalHyperparameter(
                "with_scaling", [True, False], default=True),
            TuplePairHyperparameter(
                "quantile_range", [
                    UniformFloatHyperparameter(
                        "q_min", 0.001, 0.3, default=0.001),
                    UniformFloatHyperparameter(
                        "q_max", 0.7, 0.999, default=0.7)
                ], default=(0.25, 0.75))
        ])


def normalizer():
    """Create a normalizer component."""
    return AlgorithmComponent(
        name="Normalizer",
        component_class=Normalizer,
        component_type=AlgorithmType.RESCALER,
        hyperparameters=[
            CategoricalHyperparameter(
                "norm", ["l1", "l2", "max"], default="l2")
        ])



"""Beam search implementation.

TODO:
- incorporate diversity into the rank score so that samples are differet.
"""

import torch
from torch.autograd import Variable
from copy import copy

from .rnn_code_generator import (
    load_model, create_metafeature_tensor, create_input_tensor,
    sos_index, sos_token, eos_index, eos_token, characters)


START_SAMPLE = ""


class BeamSearchDecoder(object):
    """Object to handle beam search logic."""

    def __init__(self, rnn, top_n=3):
        """Initialize BeamSearch object.

        :param CodeGeneratorRNN rnn: an RNN object.
        :param int top_n: number of candidate sequences to consider.
        """
        self.rnn = rnn
        self.top_n = top_n
        self.top_n_expand = min(top_n, rnn.output_size)

    def sample(
            self, start_char=sos_token,
            metafeatures=["executable", "creates_estimator"],
            max_length=100):
        """Sample from generator given a starting character.

        :param str start_char: starting character, default: sos_token "^"
        :param list[str] metafeatures: categorical values specified by
            rnn_code_generate.metafeature_categories.
        """
        metafeature_tensor = Variable(
            create_metafeature_tensor(metafeatures, start_char))
        hidden = self.rnn.initHidden()
        start_candidate = Candidate(0, [sos_index], start_char, hidden)
        candidates = self._expand_candidate(
            start_candidate, metafeature_tensor)

        for _ in range(max_length):
            new_candidates = []
            for c in candidates:
                new_candidates.extend(
                    self._expand_candidate(c, metafeature_tensor))
            candidates = self._prune_candidates(new_candidates)
        return candidates

    def _expand_candidate(self, candidate, metafeature_tensor):
        """Evaluates a candidate sequence."""
        # early return if the last char in sequence is eos token.
        if candidate.sequence[-1] == eos_token:
            return [candidate]
        output, hidden = self.rnn(
            metafeature_tensor,
            Variable(create_input_tensor(candidate.sequence[-1])),
            candidate.hidden_state)
        # rnn uses LogSoftmax as output layer, so no need to compute log
        # probability here.
        top_p, top_i = output.data.topk(self.top_n_expand)
        top_p, top_i = top_p.view(-1).tolist(), top_i.view(-1).tolist()
        prob_scores, indices, sequences, hidden_copies = [], [], [], []
        for p, i in zip(top_p, top_i):
            prob_scores.append(candidate.prob_score + p)
            indices.append(candidate.indices + [i])
            sequences.append(candidate.sequence + characters[i])
            hidden_copies.append(hidden.clone())
        return [Candidate(*args) for args in
                zip(prob_scores, indices, sequences, hidden_copies)]

    def _prune_candidates(self, candidates):
        """Picks the top n from a list of candidates."""
        scores = torch.Tensor([c.prob_score for c in candidates])
        _, top_cand_index = scores.topk(self.top_n)
        return [c for i, c in enumerate(candidates) if i in top_cand_index]


class Candidate(object):

    def __init__(self, prob_score, indices, sequence, hidden_state):
        """Initialize a candidate sample used by the BeamSearch routine."""
        self.prob_score = prob_score
        self.indices = indices
        self.sequence = sequence
        self.hidden_state = hidden_state

    def normalized_prob_score(self):
        pass

    @property
    def sample_code(self):
        """Clean up sample code sequence."""
        code = copy(self.sequence)
        if code.startswith(sos_token):
            code = code[1:]
        if code.endswith(eos_token):
            code = code[:-1]
        return code



"""Data environments module."""

from collections import ChainMap
from functools import partial

from ..data_types import DataSourceType
from . import openml_api, kaggle_api, sklearn_classification, \
    sklearn_regression


def get_envs_from_source(
        env_fns, dataset_names, n=None, test_size=None, random_state=None,
        verbose=None):
    envs = ChainMap(*[
        env_fn(n=n, test_size=test_size, random_state=random_state,
               verbose=verbose)
        for env_fn in env_fns])
    if dataset_names is None:
        return [envs[d] for d in envs]
    return [envs[d] for d in dataset_names if d in envs]


ENV_SOURCES = {
    # TODO: these wrappers should take a list of dataset names as arguments
    # and return only those dataset envs. Make this efficient by only creating
    # the data envs specified.
    DataSourceType.SKLEARN: partial(
        get_envs_from_source,
        [sklearn_classification.envs,
         sklearn_regression.envs]),
    DataSourceType.OPEN_ML: partial(
        get_envs_from_source,
        [openml_api.classification_envs,
         openml_api.regression_envs]),
    DataSourceType.KAGGLE: partial(
        get_envs_from_source,
        [kaggle_api.classification_envs,
         kaggle_api.regression_envs]),
    DataSourceType.AUTOSKLEARN_BENCHMARK: partial(
        get_envs_from_source,
        [openml_api.autosklearn_paper_classification_envs])
}


def envs(dataset_names=None, sources=None, target_types=None,
         test_set_config=None, env_source_map=ENV_SOURCES):
    """Get environments.

    :param list[DataSourceType] sources: only get data envs from these sources.
    :param list[str] dataset_names: only include these datasets in data dist.
    :param list[TargetType] target_types: only include these target types
        in data dist.
    :param dict[DataSourceType -> dict] test_set_config: a dictionary where
        keys DataSourceTypes and values are dictionaries of the form:
        {"test_size": float, "random_state": int}. This is used to set the
        task environment test set sizes.
    """
    # TODO: need to set aside test set for each data env. This should be
    # standardized (set a random seed) so that results are reproducible.
    test_set_config = {} if test_set_config is None else test_set_config
    _envs = []
    if sources is None:
        sources = list(env_source_map.keys())
    for env_source in sources:
        _envs.extend(env_source_map[env_source](
            dataset_names, **test_set_config.get(env_source, {})))
    if target_types:
        _envs = [e for e in _envs if e.target_type in target_types]
    return _envs



"""Evaluate a trained controller model."""

from collections import namedtuple
from sklearn.exceptions import NotFittedError
from torch.autograd import Variable

from .. import utils
from ..data_types import DataSourceType


TestSetResult = namedtuple(
    "TestSetResult", ["ml_framework", "reward", "test_score"])
InferenceResult = namedtuple(
    "InferenceResult",
    ["mlf_description", "reward", "action", "validation_score", "is_valid"])


class CASHInference(object):

    def __init__(self, controller, task_env):
        self.controller = utils.freeze_model(controller)
        self.t_env = task_env
        self._validation_scores = []
        self._successful_mlfs = []

    def evaluate_test_sets(self, dataset, mlfs):
        """Evaluate controller based on training dataenv test performance.

        Evaluation is based on the best mlfs that were found during training
        per data env.

        Note that for Kaggle data environments, the evaluation process is a
        little bit more involved since predictions need to be sent up to the
        kaggle platform to be evaluated

        TODO: should the controller be able to propose subsequent MLFs
        based on the reward of previous MLFs? Seems like this is introducing
        bias, but not sure exactly how.

        :param dict[str -> list] mlfs: key is dataset name and value is a list
            of sklearn.Pipeline objects.
        """
        inference_results = []
        data_env = [d for d in self.t_env.data_distribution
                    if d.name == dataset]
        if len(data_env) == 0:
            raise ValueError(
                "%s not in the task environment data distribution." %
                dataset)
        data_env = data_env[0]
        self.t_env.set_data_env(data_env)
        for mlf in mlfs:
            if data_env.source is DataSourceType.KAGGLE:
                if self.t_env.current_data_env.y_test is None:
                    # TODO: send a submission and get score via kaggle api.
                    print("%s is a kaggle data environment" % data_env.name)
                    reward, test_score = None, None
                else:
                    print(
                        "%s is a kaggle data environment using a holdout "
                        "partition training data to product test data. Note "
                        "this mode is for 'development', where the kaggle "
                        "test labels should only be accessible through the "
                        "kaggle API. For 'production' mode, specify a task "
                        "environment where test_set_config for KAGGLE data "
                        "source doesn't specify `test_size` or `random_state`"
                        % dataset)
            mlf, reward, test_score = self.t_env.score(
                mlf,
                self.t_env.current_data_env.X_test,
                self.t_env.current_data_env.y_test)
            if (mlf, reward, test_score) == (None, None, None):
                raise NotFittedError(
                    "%s is not a fitted MLF. Make sure that the experiment "
                    "serializes the fitted mlf pipeline." % mlf)
            inference_results.append(TestSetResult(mlf, reward, test_score))

        return inference_results

    def evaluate_training_data_envs(self, n=5, datasets=None, verbose=False):
        train_data_env_results = {}
        for train_data_env in self.t_env.data_distribution:
            if datasets and train_data_env.name not in datasets:
                continue
            self.t_env.set_data_env(train_data_env)
            train_data_env_results[train_data_env.name] = self.infer(
                n, verbose)
        return train_data_env_results

    def evaluate_test_data_envs(self, n=5, datasets=None, verbose=False):
        """Test data env evaluation.

        This workflow evaluates the controller with respect to a new dataset
        that it hasn't seen before. The new dataset has three partitions:
        training, validation, and test set. The controller should then propose
        MLFs, measured against the validation set. Since the controller is
        hypothesized to implement a meta-RL algorithm, it should be able to
        learn even if all the controller's weights are frozen.
        """
        test_data_env_results = {}
        for test_data_env in self.t_env.test_data_distribution:
            if datasets and test_data_env.name not in datasets:
                continue
            print("evaluating test data env: %s" % test_data_env.name)
            self.t_env.set_data_env(test_data_env)
            test_data_env_results[test_data_env.name] = self.infer(
                n, verbose)
        return test_data_env_results

    def infer(self, n, verbose):
        """Make inferences by sampling the task environment.

        :param int n: number of times to sample task environment.
        :param bool verbose: prints metric logs during training.
        :returns list[InferenceResult]: a named tuple
        """
        prev_reward, prev_action = 0, self.controller.init_action()
        inference_results = []
        n_valid_mlf = 0
        for i in range(n):
            mlf, inference = self._infer_iter(
                self.t_env.sample_task_state(data_env_partition="test"),
                self.t_env.current_data_env.target_type,
                prev_reward,
                prev_action)
            # TODO: serialize the top k mlfs here
            inference_results.append(inference)
            prev_reward, prev_action = inference.reward, inference.action
            n_valid_mlf += int(inference.is_valid)
            if verbose:
                print(
                    "iter %d - n valid mlf: %d/%d%s" % (
                        i, n_valid_mlf, i + 1, " " * 10),
                    sep=" ", end="\r", flush=True)
        return inference_results

    def _infer_iter(
            self, metafeature_tensor, target_type, prev_reward, prev_action):
        mlf, action_activation = self.propose_mlf(
            metafeature_tensor, target_type, prev_reward, prev_action)
        mlf, reward, validation_score, is_valid = self.evaluate_mlf(mlf)

        return mlf, InferenceResult(
            str(mlf.named_steps) if mlf is not None else mlf,
            reward, Variable(action_activation),
            validation_score, is_valid)

    def propose_mlf(
            self, task_state_tensor, target_type, prev_reward, prev_action):
        """Given a task state,  propose a machine learning framework."""
        actions, action_activation = self.controller.decode(
            init_input_tensor=prev_action,
            target_type=target_type,
            aux=utils.aux_tensor(prev_reward),
            metafeatures=Variable(task_state_tensor),
            init_hidden=self.controller.init_hidden())
        algorithms, hyperparameters = utils.get_mlf_components(actions)
        mlf = self.controller.a_space.create_ml_framework(
            algorithms, hyperparameters=hyperparameters,
            task_metadata=self.t_env.get_current_task_metadata())
        return mlf, action_activation.data

    def evaluate_mlf(self, mlf):
        """Evaluate actions on the validation set of the data environment."""
        mlf, reward, validation_score = self.t_env.evaluate(mlf)
        reward, is_valid = self.is_valid_mlf(reward)
        return mlf, reward, validation_score, is_valid

    def is_valid_mlf(self, reward):
        if reward is None:
            return self.t_env.error_reward, False
        return reward,  True



"""Module for generating samples from a pre-trained RNN.

TODO:
- measure proportion of samples are: executable, evalutes to estimator object.
"""

from rnn_code_generator import load_model, generate_samples


def main(path="rnn_code_generator_model.pt"):
    """Generate samples from the rnn_code_generator.py module."""
    rnn = load_model(path)
    print("samples: executable, creates_estimator")
    for _ in range(5):
        print(generate_samples(
            rnn, start_chars="^",
            metafeatures=[["executable", "creates_estimator"]]))
    print("\nsamples: not_executable, not_creates_estimator")
    for _ in range(5):
        print(generate_samples(
            rnn, start_chars="^",
            metafeatures=[["not_executable", "not_creates_estimator"]]))


if __name__ == "__main__":
    main()



"""Source datasets from OpenML.

https://www.openml.org/search?type=data
"""

import logging

from collections import OrderedDict
from functools import partial

import numpy as np
import openml

from openml.exceptions import OpenMLServerException

from .data_environment import DataEnvironment
from . import autosklearn_clf_task_ids
from ..data_types import FeatureType, TargetType, OpenMLTaskType, \
    DataSourceType


logger = logging.getLogger(__name__)


N_CLASSIFICATION_ENVS = 20
N_REGRESSION_ENVS = 20
SPARSE_DATA_FORMAT = "Sparse_ARFF"


# Map OpenML feature data types to metalearn feature data types
FEATURE_TYPE_MAP = {
    "nominal": FeatureType.CATEGORICAL,
    "numeric": FeatureType.CONTINUOUS,
}
DATASOURCE_TYPES = [
    DataSourceType.OPEN_ML, DataSourceType.AUTOSKLEARN_BENCHMARK]


def _map_feature(feature):
    feature_dtype = FEATURE_TYPE_MAP.get(feature.data_type, None)
    if feature_dtype is None:
        raise ValueError(
            "feature dtype %s not recognized. valid values are %s" % (
                FEATURE_TYPE_MAP.keys()))
    return feature_dtype


def _normalize_feature_indices(feature_indices):
    min_index = min(feature_indices)
    return [i - min_index for i in feature_indices]


def openml_to_data_env(
        openml_dataset, target_column, target_type,
        data_source_type, test_size, random_state, target_preprocessor=None,
        include_features_with_na=False):
    print("fetching openml dataset: %s" % openml_dataset.name)
    if data_source_type not in DATASOURCE_TYPES:
        raise ValueError(
            "expected one of %s, found : %s" % (
                DATASOURCE_TYPES, data_source_type))
    feature_indices = []
    feature_types = []
    target_index = None
    ignore_atts = openml_dataset.ignore_attributes
    row_id_att = openml_dataset.row_id_attribute
    for key, feature in openml_dataset.features.items():
        if (ignore_atts and feature.name in ignore_atts) or \
                (row_id_att and feature.name == row_id_att):
            print("ignoring attribute %s in dataset %s" %
                  (ignore_atts, openml_dataset.name))
            continue

        if feature.name == target_column:
            target_index = key
            continue
        if feature.number_missing_values != 0:
            if include_features_with_na:
                feature_indices.append(key)
                feature_types.append(_map_feature(feature))
            else:
                continue
        else:
            feature_indices.append(key)
            feature_types.append(_map_feature(feature))

    if target_index is None:
        print("target column %s not found for dataset %s, skipping." %
              (target_column, openml_dataset.name))
        return None
    if len(feature_indices) == 0:
        print("no features found for dataset %s, skipping." %
              openml_dataset.name)
        return None

    def _fetch_training_data():
        # see https://openml.github.io/openml-python/develop/generated/openml.OpenMLDataset.html#openml.OpenMLDataset  # noqa E53
        # for more details
        data, *_ = openml_dataset.get_data(
            include_row_id=True, include_ignore_attributes=True,
            dataset_format="array")
        if openml_dataset.format == SPARSE_DATA_FORMAT:
            # TODO: currently the controller can't handle sparse matrices.
            data = data.todense()

        # remove rows with null targets. Only case we'll need that is if
        # metalearn supports semi-supervised learning.
        X = data[:, feature_indices]
        y = data[:, target_index].ravel()
        y_not_nan = ~np.isnan(y)
        return X[y_not_nan, :], y[y_not_nan]

    # normalize feature indices so that index is aligned with
    # _fetch_training_data
    min_index = min(feature_indices)
    normalized_feature_indices = [i - min_index for i in feature_indices]

    return DataEnvironment(
        name="openml.%s" % openml_dataset.name.lower(),
        source=data_source_type,
        target_type=target_type,
        feature_types=feature_types,
        feature_indices=normalized_feature_indices,
        # include row id ensures that indices are correctly aligned
        fetch_training_data=_fetch_training_data,
        fetch_test_data=None,
        test_size=test_size,
        random_state=random_state,
        target_preprocessor=target_preprocessor,
        scorer=None,
    )


def _get_target_type(num_classes, task_type):
    if task_type is OpenMLTaskType.SUPERVISED_REGRESSION:
        return TargetType.REGRESSION
    elif task_type is OpenMLTaskType.SUPERVISED_CLASSIFICATION:
        if num_classes > 2:
            return TargetType.MULTICLASS
        else:
            return TargetType.BINARY
    else:
        raise ValueError(
            "task_type %s is currently not recognized by metalearn." %
            task_type)


def _get_dataset_metadata(openml_task_metadata, task_type):
    target_columns, target_types = zip(*[
        (v["target_feature"],
         _get_target_type(v.get("NumberOfClasses", 0), task_type))
        for v in openml_task_metadata.values()])
    dataset_ids = [v["source_data"] for v in openml_task_metadata.values()]
    datasets, target_columns_out, target_types_out = [], [], []
    for id, tcols, ttypes in zip(dataset_ids, target_columns, target_types):
        try:
            datasets.append(openml.datasets.get_dataset(id))
            target_columns_out.append(tcols)
            target_types_out.append(ttypes)
        except OpenMLServerException as e:
            print("error when fetching openml metadata: %s, skipping dataset"
                  % e)
    return datasets, target_columns_out, target_types_out


def _create_envs(
        data_source_type, datasets, target_columns, target_types, test_size,
        random_state):
    _openml_to_data_env = partial(
        openml_to_data_env,
        data_source_type=data_source_type,
        test_size=test_size,
        random_state=random_state,
        include_features_with_na=True)
    envs = [_openml_to_data_env(*args) for args in zip(
                datasets, target_columns, target_types)]
    return OrderedDict([(d.name, d) for d in envs if d])


def classification_envs(
        n=N_CLASSIFICATION_ENVS, test_size=None, random_state=None,
        verbose=False):
    if n is None:
        n = N_CLASSIFICATION_ENVS
    task_type = OpenMLTaskType.SUPERVISED_CLASSIFICATION
    dataset_metadata = _get_dataset_metadata(
        openml.tasks.list_tasks(task_type_id=task_type.value, size=n),
        task_type)
    return _create_envs(
        DataSourceType.OPEN_ML, *dataset_metadata +
        (test_size, random_state))


def regression_envs(
        n=N_REGRESSION_ENVS, test_size=None, random_state=None, verbose=False):
    if n is None:
        n = N_CLASSIFICATION_ENVS
    task_type = OpenMLTaskType.SUPERVISED_REGRESSION
    dataset_metadata = _get_dataset_metadata(
        openml.tasks.list_tasks(task_type_id=task_type.value, size=n),
        task_type)
    return _create_envs(
        DataSourceType.OPEN_ML, *dataset_metadata + (test_size, random_state))


def _task_ids(n):
    if n is None:
        task_ids = autosklearn_clf_task_ids.TASK_IDS
    else:
        task_ids = autosklearn_clf_task_ids.TASK_IDS[:n]
    return task_ids


def _get_task(id, verbose):
    if verbose:
        print("getting task metadata for task id %d" % id)
    try:
        task = openml.tasks.get_task(id)
        if task is None:
            print("skipping dataset %d id. "
                  "openml.tasks.get_task return null" % id)
        return task
    except OpenMLServerException as e:
        print("skipping dataset %d. error: %s" % (id, e))


def autosklearn_paper_classification_envs(
        n=None, test_size=None, random_state=None, verbose=False):
    """From Feurer et al. "Efficient and Robust Automated Machine Learning.

    https://papers.nips.cc/paper/5872-efficient-and-robust-automated-machine-learning  # noqa
    """
    if n is None:
        task_ids = autosklearn_clf_task_ids.TASK_IDS
    else:
        task_ids = autosklearn_clf_task_ids.TASK_IDS[:n]
    openml_tasks = list(filter(
        None, [_get_task(id, verbose) for id in task_ids]))
    openml_datasets = [t.get_dataset() for t in openml_tasks]
    target_columns = (t.target_name for t in openml_tasks)
    target_types = (_get_target_type(
        len(t.class_labels),
        OpenMLTaskType.SUPERVISED_CLASSIFICATION) for t in openml_tasks)
    return _create_envs(
        DataSourceType.AUTOSKLEARN_BENCHMARK, openml_datasets, target_columns,
        target_types, test_size, random_state)



"""Module for creating a structured algorithm environment.

The structured algorithm space is different from the `algorithm_env.py`
module because here we specify an interface to generate a machine learning
framework F, which is a sequence of algorithms (estimators/transformers) and
their associated hyperparameters. This interface is inspired by this paper
on automated machine learning:

papers.nips.cc/paper/5872-efficient-and-robust-automated-machine-learning.pdf

The sequence follows the general structure

.-------------------.    .----------------------.    .----------------------.
| data preprocessor | -> | feature preprocessor | -> | classifier/regressor |
.-------------------.    .----------------------.    .----------------------.

where there can be n data preprocessors (imputation, scaling, encoding,
filtering).
"""

import numpy as np

from collections import OrderedDict

from sklearn.pipeline import Pipeline

from . import components
from .data_types import AlgorithmType, TargetType
from .components.constants import START_TOKEN, END_TOKEN, NONE_TOKEN


SPECIAL_TOKENS = [START_TOKEN, END_TOKEN, NONE_TOKEN]
# ml framework pipeline must have this signature. Can eventually support
# multiple signatures.
CLASSIFIER_MLF_SIGNATURE = [
    AlgorithmType.IMPUTER,
    AlgorithmType.ONE_HOT_ENCODER,
    AlgorithmType.RESCALER,
    AlgorithmType.FEATURE_PREPROCESSOR,
    AlgorithmType.CLASSIFIER
]
REGRESSOR_MLF_SIGNATURE = [
    AlgorithmType.IMPUTER,
    AlgorithmType.ONE_HOT_ENCODER,
    AlgorithmType.RESCALER,
    AlgorithmType.FEATURE_PREPROCESSOR,
    AlgorithmType.REGRESSOR
]
TARGET_TYPE_TO_MLF_SIGNATURE = {
    TargetType.BINARY: CLASSIFIER_MLF_SIGNATURE,
    TargetType.MULTICLASS: CLASSIFIER_MLF_SIGNATURE,
    TargetType.REGRESSION: REGRESSOR_MLF_SIGNATURE,
    TargetType.MULTIREGRESSION: REGRESSOR_MLF_SIGNATURE,
}


class AlgorithmSpace(object):
    """Class that generates machine learning frameworks."""

    ALL_COMPONENTS = [
        AlgorithmType.IMPUTER,
        AlgorithmType.ONE_HOT_ENCODER,
        AlgorithmType.RESCALER,
        AlgorithmType.FEATURE_PREPROCESSOR,
        AlgorithmType.CLASSIFIER,
        AlgorithmType.REGRESSOR
    ]

    def __init__(self,
                 data_preprocessors=None,
                 feature_preprocessors=None,
                 classifiers=None,
                 regressors=None,
                 with_start_token=False,
                 with_end_token=False,
                 with_none_token=False,
                 hyperparam_with_start_token=False,
                 hyperparam_with_end_token=False,
                 hyperparam_with_none_token=False,
                 random_state=None):
        """Initialize a structured algorithm environment.

        :param list[AlgorithmComponent]|None data_preprocessors: algorithm
            components for data preprocessing
        :param list[AlgorithmComponent]|None feature_preprocessors: algorithm
            components for feature preprocessing
        :param list[AlgorithmComponent]|None classifiers: algorithm
            components for classification
        """
        self.data_preprocessors = get_data_preprocessors() \
            if data_preprocessors is None else data_preprocessors
        self.feature_preprocessors = get_feature_preprocessors() \
            if feature_preprocessors is None else feature_preprocessors
        self.classifiers = get_classifiers() if classifiers is None \
            else classifiers
        self.regressors = get_regressors() if regressors is None \
            else regressors
        # TODO: assess whether these tokens are necessary
        self.with_start_token = with_start_token
        self.with_end_token = with_end_token
        self.with_none_token = with_none_token
        self.hyperparam_with_start_token = hyperparam_with_start_token
        self.hyperparam_with_end_token = hyperparam_with_end_token
        self.hyperparam_with_none_token = hyperparam_with_none_token
        self.random_state = random_state
        np.random.seed(self.random_state)

    @property
    def components(self):
        """Concatenate all components into a single list."""
        components = self.data_preprocessors + \
            self.feature_preprocessors + \
            self.classifiers + \
            self.regressors
        if self.with_start_token:
            components += [START_TOKEN]
        if self.with_end_token:
            components += [END_TOKEN]
        if self.with_none_token:
            components += [NONE_TOKEN]
        return components

    @property
    def n_components(self):
        """Return number of components in the algorithm space."""
        return len(self.components)

    def component_dict_from_signature(self, signature):
        """Return dictionary of algorithm types and list of components.

        :param list[str] signature: ML framework signature indicating the
            ordering of algorithm components to form a sklearn Pipeline.
        """
        return OrderedDict([
            (component_type, self.get_components(component_type))
            for component_type in signature])

    def component_dict_from_target_type(self, target_type):
        """Get algorithm components based on target type.

        :param data_types.TargetType target_type: get the MLF signature for
            this target.
        """
        return self.component_dict_from_signature(
            TARGET_TYPE_TO_MLF_SIGNATURE[target_type])

    def get_components(self, component_type):
        """Get all components of a particular type.

        :param str component_type: type of algorithm
        :returns: list of components of `component_type`
        :rtype: list[AlgorithmComponent]
        """
        return [c for c in self.components if c not in SPECIAL_TOKENS and
                c.component_type == component_type]

    def h_state_space(self, components, with_none_token=False):
        """Get hyperparameter state space by components.

        :param list[AlgorithmComponent] components: list of components
        :returns: OrderedDict of hyperparameters
        :rtype: dict[str -> list]
        """
        hyperparam_states = OrderedDict()
        for c in components:
            if c not in SPECIAL_TOKENS and c.hyperparameters is not None:
                hyperparam_states.update(
                    c.hyperparameter_state_space(with_none_token))
        return hyperparam_states

    def h_exclusion_conditions(self, components):
        """Get the conditional map of which hyperparameters go together."""
        exclude_conditions = OrderedDict()
        for c in components:
            if c and c.hyperparameters is not None:
                exclude_conditions.update(
                    c.hyperparameter_exclusion_conditions())
        return exclude_conditions

    def sample_component(self, component_type):
        """Sample a component of a particular type.

        :param str component_type: type of algorithm, one of
            {"one_hot_encoder", "imputer", "rescaler", "feature_preprocessor",
             "classifier", "regressor"}
        :returns: a sampled algorithm component of type `component_type`.
        :rtype: AlgorithmComponent
        """
        component_subset = self.get_components(component_type)
        return component_subset[np.random.randint(len(component_subset))]

    def sample_components_from_signature(self, signature):
        """Sample algorithm components from ML signature.

        :param list[str] signature: ML framework signature indicating the
            ordering of algorithm components to form a sklearn Pipeline.
        """
        return [self.sample_component(component_type)
                for component_type in signature]

    def sample_ml_framework(self, signature, task_metadata=None):
        """Sample a random ML framework from the algorithm space.

        :param list[str] signature: ML framework signature indicating the
            ordering of algorithm components to form a sklearn Pipeline.
        :param int|None random_state: provide random state, which determines
            the ML framework sampled.
        :param dict[str, any] task_metadata: constraints imposed by the
            environment on the hyperparameter space.
        """
        components = self.sample_components_from_signature(signature)
        framework_hyperparameters = {}
        for a in components:
            framework_hyperparameters.update(
                a.sample_hyperparameter_state_space())
        return self.create_ml_framework(
            components, hyperparameters=framework_hyperparameters,
            task_metadata=task_metadata)

    def create_ml_framework(
            self, components, memory=None, hyperparameters=None,
            task_metadata=None):
        """Create ML framework, in this context an sklearn pipeline object.

        :param list[AlgorithmComponent] components: A list of algorithm
            components with which to create an ML framework.
        :param str|None memory: path to caching directory in which to store
            fitten transformers of the sklearn.Pipeline. If None, no caching
            is done
        :param dict[str, str] hyperparameters: picked by the automl system.
        :param dict[str, any] task_metadata: constraints imposed by the
            environment on the hyperparameter space.
        """
        steps = []
        hyperparameters = {} if hyperparameters is None else hyperparameters
        task_metadata = {} if task_metadata is None else task_metadata
        for component in components:
            steps.append(
                (component.name, component(
                    categorical_features=task_metadata.get(
                        "categorical_features", None),
                    continuous_features=task_metadata.get(
                        "continuous_features", None)
                    ))
            )
            hyperparameters.update(component.get_constant_hyperparameters())
        ml_framework = Pipeline(memory=memory, steps=steps)
        ml_framework.set_params(**hyperparameters)
        return ml_framework

    def _combine_dicts(self, dicts):
        combined_dicts = {}
        for d in dicts:
            for h in d:
                combined_dicts.update(h)
        return combined_dicts

    def set_random_state(self, random_state):
        np.random.seed(random_state)

    @property
    def config(self):
        return {
            "data_preprocessors": self.data_preprocessors,
            "feature_preprocessors": self.feature_preprocessors,
            "classifiers": self.classifiers,
            "regressors": self.regressors,
            "with_start_token": self.with_start_token,
            "with_end_token": self.with_end_token,
            "with_none_token": self.with_none_token,
            "hyperparam_with_start_token": self.hyperparam_with_start_token,
            "hyperparam_with_end_token": self.hyperparam_with_end_token,
            "hyperparam_with_none_token": self.hyperparam_with_none_token,
            "random_state": self.random_state,
        }

    def __eq__(self, other):
        return self.config == other.config


def get_data_preprocessors():
    """Get all data preprocessors in structured algorithm space."""
    return [
        components.data_preprocessors.simple_imputer(),
        components.data_preprocessors.one_hot_encoder(),
        components.data_preprocessors.minmax_scaler(),
        components.data_preprocessors.standard_scaler(),
        components.data_preprocessors.robust_scaler(),
        components.data_preprocessors.normalizer(),
    ]


def get_feature_preprocessors():
    """Get all feature preprocessors in structured algorithm space."""
    return [
        components.feature_preprocessors.fast_ica(),
        components.feature_preprocessors.feature_agglomeration(),
        components.feature_preprocessors.kernel_pca(),
        components.feature_preprocessors.rbf_sampler(),
        components.feature_preprocessors.nystroem_sampler(),
        components.feature_preprocessors.pca(),
        components.feature_preprocessors.polynomial_features(),
        components.feature_preprocessors.random_trees_embedding(),
        components.feature_preprocessors.truncated_svd(),
        components.feature_preprocessors.variance_threshold_filter(),
    ]


def get_classifiers():
    """Get all classifiers in structured algorithm space."""
    return [
        components.classifiers.adaboost(),
        components.classifiers.decision_tree(),
        components.classifiers.gaussian_naive_bayes(),
        components.classifiers.gradient_boosting(),
        components.classifiers.k_nearest_neighbors(),
        components.classifiers.logistic_regression(),
        components.classifiers.multinomial_naive_bayes(),
        components.classifiers.random_forest_classifier(),
        components.classifiers.rbf_gaussian_process_classifier(),
        components.classifiers.support_vector_classifier_linear(),
        components.classifiers.support_vector_classifier_nonlinear(),
    ]


def get_regressors():
    """Get all classifiers in structured algorithm space."""
    return [
        components.regressors.adaboost_regression(),
        components.regressors.ard_regression(),
        components.regressors.bayesian_ridge_regression(),
        components.regressors.decision_tree_regression(),
        components.regressors.k_nearest_neighbors_regression(),
        components.regressors.lasso_regression(),
        components.regressors.random_forest_regression(),
        components.regressors.rbf_gaussian_process_regression(),
        components.regressors.ridge_regression(),
        components.regressors.support_vector_regression_linear(),
        components.regressors.support_vector_regression_nonlinear(),
    ]
